----------------------------------------------------------------------------------------------------
-- Random Tour Lua Script
-- By Scepheo
----------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------
-- Script configuration
----------------------------------------------------------------------------------------------------

-- How big the checkpoint should be in relation to the node
local nodeSizeMultiplier = 1.5

-- Distance after which a new checkpoint should be placed
local nodeMaxDistance = 50

-- Distance after which the first checkpoint should be placed
local nodeInitialDistance = 10

-- How many nodes (not checkpoints!) the path needs to contain
local minPathLength = 15

-- How sharp a corner needs to be (dot product) to place a waypoint after it
local angleThreshold = 0.9

-- Maximum distance between any two waypoints
local distanceThreshold = 150

-- Distance at which to draw debugging information
local debugDrawDistance = 100

-- Max distance for an 'interesting' waypoint to be from the first 'interesting' waypoint in the set
local maxInterestDistance = 25


----------------------------------------------------------------------------------------------------
-- Global variables
----------------------------------------------------------------------------------------------------

local graphpath = require('graphpath')
local settings = require('scenario/random_tour_settings')
math.randomseed(os.time())

local loaded = false
local running = false
local mapInstance = {}
local route

local selectedVehicle


----------------------------------------------------------------------------------------------------
-- Helper functions
----------------------------------------------------------------------------------------------------

local waypointNumber = 1

local function addWaypointFor(scenario, routeEntry)
	local position = routeEntry.pos.x .. ' ' .. routeEntry.pos.y .. ' ' .. routeEntry.pos.z
	local wpName = 'random_tour_wp_' .. tostring(waypointNumber)
	waypointNumber = waypointNumber + 1

	TorqueScript.eval([[
		if (isObject(]] .. wpName .. [[))
		{
			]] .. wpName .. [[.delete();
		}
		
		new BeamNGWaypoint(]] .. wpName .. [[)
		{
			position = "]] .. position .. [[";
		};
		
		ScenarioObjectsGroup.add(]] .. wpName .. [[);
	]])
	
	table.insert(scenario.lapConfig, wpName)
	
	scenario.nodes[wpName] = 
	{
		pos = vec3(routeEntry.pos),
		radius = routeEntry.radius * nodeSizeMultiplier
	}
end

local function getDistanceBetween(aNode, bNode)
	if not aNode then return 0 end
	
	local a = mapInstance.nodes[aNode].pos
	local b = mapInstance.nodes[bNode].pos
	
	return (a - b):length()
end

local function getAngleBetween(aNode, bNode, cNode)
	if not aNode or not cNode then return 1 end
	
	local a = mapInstance.nodes[aNode].pos
	local b = mapInstance.nodes[bNode].pos
	local c = mapInstance.nodes[cNode].pos
	
	local x1 = b.x - a.x
	local y1 = b.y - a.y
	local l1 = math.sqrt(x1 * x1 + y1 * y1)
	x1 = x1 / l1
	y1 = y1 / l1
	
	local x2 = c.x - b.x
	local y2 = c.y - b.y
	local l2 = math.sqrt(x2 * x2 + y2 * y2)
	x2 = x2 / l2
	y2 = y2 / l2
	
	return x1 * x2 + y1 * y2
end


----------------------------------------------------------------------------------------------------
-- Time of Day
----------------------------------------------------------------------------------------------------

local todObject
local duskStart
local nightStart
local dawnStart
local dayStart

local function initTimeOfDay(scenario)
	local todList = scenetree.findClassObjects('TimeOfDay')
	
	if #todList == 1 then
		todObject = scenetree.findObject(todList[1])
	end
	
	duskStart = scenario.duskStart
	nightStart = scenario.nightStart
	dawnStart = scenario.dawnStart
	dayStart = scenario.dayStart
end

local function isBetween(low, high, value)
	if low < high then
		return value >= low and value < high
	else
		return value >= low or value < high
	end
end

local function setVehicleLights()
	if not settings.setRandomTime or not todObject then
		return
	end
	
	local newTime = todObject.time
	local lightMode
	
	if isBetween(dawnStart, dayStart, newTime) then
		lightMode = '1'
	elseif isBetween(dayStart, duskStart, newTime) then
		lightMode = '0'
	elseif isBetween(duskStart, nightStart, newTime) then
		lightMode = '1'
	else
		lightMode = '2'
	end
	
	scenetree.scenario_player0:queueLuaCommand('electrics.setLightsState('..lightMode..')')
end

local function randomTimeBetween(low, high)
	return (((1 + high - low) % 1) * math.random() + low) % 1
end

local function setRandomTime()
	if not settings.setRandomTime or not todObject then
		return
	end

	-- select dawn, day, dusk or night
	local partOfDay = math.random()
	local newTime
	
	if partOfDay < 0.2 then
		log('I', 'random_tour', 'setRandomTime: Selected dawn')
		newTime = randomTimeBetween(dawnStart, dayStart)
	elseif partOfDay < 0.6 then
		log('I', 'random_tour', 'setRandomTime: Selected day')
		newTime = randomTimeBetween(dayStart, duskStart)
	elseif partOfDay < 0.8 then
		log('I', 'random_tour', 'setRandomTime: Selected dusk')
		newTime = randomTimeBetween(duskStart, nightStart)
	else
		log('I', 'random_tour', 'setRandomTime: Selected night')
		newTime = randomTimeBetween(nightStart, dawnStart)
	end
	
	log('I', 'random_tour', 'setRandomTime: Selected time '..tostring(newTime))	
	todObject.time = newTime
	todObject.play = true
	todObject.dayLength = 3600
	todObject.dayScale = 1
	todObject.nightScale = 1
end


----------------------------------------------------------------------------------------------------
-- Random vehicle selection
----------------------------------------------------------------------------------------------------

local function hue2rgb(p, q, t)
	if t < 0   then t = t + 1 end
	if t > 1   then t = t - 1 end
	if t < 1/6 then return p + (q - p) * 6 * t end
	if t < 1/2 then return q end
	if t < 2/3 then return p + (q - p) * (2/3 - t) * 6 end
	return p
end

local function hslToRgb(h, s, l)
	local r, g, b

	if s == 0 then
		r, g, b = l, l, l
	else
		local q
		
		if l < 0.5 then
			q = l * (1 + s)
		else
			q = l + s - l * s
		end
		
		local p = 2 * l - q

		r = hue2rgb(p, q, h + 1/3)
		g = hue2rgb(p, q, h)
		b = hue2rgb(p, q, h - 1/3)
	end

	return r, g, b
end

local function loadPlayer(position, direction)
	
	-- Set the player at the starting location
	local tsPosition = string.format('"%.3f %.3f %.3f"', position.x, position.y, position.z)
	
	local angle = math.deg(math.atan2(direction.x, direction.y)) + 180
	local tsRotation = string.format('"0 0 1 %.3f"', angle)
	
	TorqueScript.eval('scenario_player0.setTransform(matrixCreate('..tsPosition..','..tsRotation..'));')
	
	local config = bnse.randomize.vehicle('scenario_player0')
	selectedVehicle = config.Name
end


----------------------------------------------------------------------------------------------------
-- Random route generation
----------------------------------------------------------------------------------------------------

local gp

local function initGraphPath()
	gp = graphpath.newGraphpath()
	
    for k, n in pairs(mapInstance.nodes) do
        n.pos = vec3(n.pos)
        gp:setPointPosition(k, n.pos)
        n.drivability = 1
    end
	
    for nid, n in pairs(mapInstance.nodes) do
        for lid, dif in pairs(n.links) do
            local totaldif = mapInstance.nodes[nid].pos:distance(mapInstance.nodes[lid].pos) * (1 / (dif + 1e-30))
            gp:bidiEdge(nid, lid, totaldif)
        end
    end
end

local function appendRandomPath(currentPath)
	
	local currentEndNode = currentPath[#currentPath]
	local penUltimateNode = currentPath[#currentPath - 1]
	
	local newEndNode = tableChooseRandomKey(mapInstance.nodes)
	local newPath = gp:getPath(currentEndNode, newEndNode)
	
	local tries = 100
	
	while tries > 0 and (newPath == nil or #newPath < 2 or newPath[2] == penUltimateNode) do
		newEndNode = tableChooseRandomKey(mapInstance.nodes)
		newPath = gp:getPath(currentEndNode, newEndNode)
		tries = tries - 1
	end
	
	if tries > 0 then
		for i = 2, #newPath do
			table.insert(currentPath, newPath[i])
		end
	end
	
	return currentPath
end

local function getRandomPath()
	initGraphPath()
	
	local startNode = tableChooseRandomKey(mapInstance.nodes)
	local path = { startNode }
	
	while #path < minPathLength do
		for i = 1, math.random(1, 8) do
			path = appendRandomPath(path)
		end
	end
	
	return path
end

local function makeRoute(path)
	local result = {}
	local distanceToStart = 0
	
	for i = 1, #path do
		local node = mapInstance.nodes[path[i]]
		local distanceToLast = getDistanceBetween(path[i - 1], path[i])
		distanceToStart = distanceToStart + distanceToLast
		
		table.insert(result, {
			index = i,
			name = path[i],
			pos = node.pos,
			radius = node.radius,
			road = string.match(path[i], "(.-)%d+$"),
			angle = getAngleBetween(path[i - 1], path[i], path[i + 1]),
			distanceToLast = distanceToLast,
			distanceToStart = distanceToStart,
			linkCount = tableSize(node.links)
		})
		
	end
	
	return result
end

local function copyRouteEntry(routeEntry)
	return {
		index = routeEntry.index,
		name = routeEntry.name,
		pos = routeEntry.pos,
		radius = routeEntry.radius,
		road = routeEntry.road,
		angle = routeEntry.angle,
		distanceToLast = routeEntry.distanceToLast,
		distanceToStart = routeEntry.distanceToStart,
		linkCount = routeEntry.linkCount
	}
end


----------------------------------------------------------------------------------------------------
-- Waypoint generation
----------------------------------------------------------------------------------------------------

local function indexOfInterest(index, recurse)
	recurse = (recurse == nil) or recurse
	
	if route[index].angle < angleThreshold then
		return true
	end
	
	if route[index].linkCount >= 2 and route[index].angle < 0.95 then
		return true
	end
	
	if recurse and index < #path then
		return indexOfInterest(index + 1, false)
	end
	
	return false
end

local function createImportantWaypoints()
	
	-- Create waypoints for corner nodes
	local i = 1
	local waypoints = {}
	
	while i < #route do
		if indexOfInterest(i) then
			local startNode = route[i]
		
			i = i + 1
			local distance = route[i].distanceToLast
						
			while i < #route - 1 and distance < maxInterestDistance and indexOfInterest(i) do
				i = i + 1
				distance = distance + route[i].distanceToLast
			end
			
			local fromNode = route[i - 1]
			local toNode = route[i]
			
			local c = fromNode.radius * 3
			local alreadyCoveredDistance = fromNode.distanceToStart - startNode.distanceToStart
			c = math.max(c - alreadyCoveredDistance, 0)
			
			local dir = toNode.pos - fromNode.pos
			local length = dir:length()
			
			local offset = Point3F(dir.x * c / length, dir.y * c / length, dir.z * c / length)
			
			local dummyRouteEntry = copyRouteEntry(fromNode)
			dummyRouteEntry.pos = fromNode.pos + offset
			
			table.insert(waypoints, dummyRouteEntry)
		else
			i = i + 1
		end
	end
		
	-- Insert last waypoint
	table.insert(waypoints, route[#route])
	
	return waypoints
end

local function fillGaps(importantWaypoints)
	local firstNode = route[1]
	local waypoints = { }
	
	for i = 1, #importantWaypoints do
		local secondNode = importantWaypoints[i]
		
		local distance = secondNode.distanceToStart - firstNode.distanceToStart
		
		if distance > distanceThreshold then
			local count = math.floor(distance / distanceThreshold)
			local stepSize = distance / (count + 1)
			
			local currentDistance = 0
			
			for j = firstNode.index + 1, secondNode.index - 1 do
				local step = route[j].distanceToStart - route[j - 1].distanceToStart
				currentDistance = currentDistance + step
				
				if currentDistance >= stepSize then
					currentDistance = currentDistance - stepSize
					table.insert(waypoints, route[j])
				end
			end
		end
		
		table.insert(waypoints, secondNode)
		firstNode = secondNode
	end
	
	return waypoints
end

----------------------------------------------------------------------------------------------------
-- Scenario events
----------------------------------------------------------------------------------------------------

local function onLoad(scenario)
	initTimeOfDay(scenario)
	setRandomTime()
	
	mapInstance = map.getMap()
	
	local path = getRandomPath()
	route = makeRoute(path)
	
	log('D', 'random_tour.onLoad', 'Using route:')
	for _, routeEntry in pairs(route) do
		log('D', 'random_tour.onLoad', string.format('    %s', routeEntry.name))
	end
	
	local waypoints = createImportantWaypoints()
	waypoints = fillGaps(waypoints)
	
	scenario.lapConfig = {}
	
	for _, routeEntry in pairs(waypoints) do
		addWaypointFor(scenario, routeEntry)
	end
	
	local startDirection
	
	for i = 1, #route do
		if route[i].distanceToStart >= nodeInitialDistance then
			startDirection = route[i].pos - route[1].pos
			break
		end
	end
	
	loadPlayer(route[1].pos, startDirection)
end

local function onScenarioChange(scenario)
	if scenario and scenario.state == 'pre-start' then
		log('I', 'random_tour.onScenarioChange', 'Scenario started, starting set-up')
		onLoad(scenario)
		log('I', 'random_tour.onScenarioChange', 'Scenario set-up complete')
	end
end

local function onPhysicsEngineEvent(eventType)
	-- Freeze whenever a vehicle is spawned
	if eventType == 'spawn' then
		log('I', 'random_tour.onPhysicsEngineEvent', 'Vehicle spawned, freezing scenario_player0')
		scenetree.scenario_player0:queueLuaCommand('scenario.freeze(1)')
		setVehicleLights()
	end
end

local function onRaceStart()
	local routeLength = route[#route].distanceToStart / 1000
	local routeDisplay = string.format('Route length: %.1f km', routeLength)
	local vehicleDisplay = string.format('Vehicle: %s', selectedVehicle)
	guihooks.trigger('ScenarioFlashMessageReset')
	guihooks.trigger('ScenarioFlashMessage', {
		{ vehicleDisplay, 5, false },
		{ routeDisplay, 5, false }
	})
end


----------------------------------------------------------------------------------------------------
-- Debugging
----------------------------------------------------------------------------------------------------

require('mathlib')
local up = vec3(0, 0, 1)
local textColor = ColorF(0, 0, 0, 0.5)
local straightLineColor = ColorF(0, 1, 0, 0.8)
local angledLineColor = ColorF(1, 0, 0, 0.8)

local function debugText(routeEntry)
	return String(string.format("%d (%.2f)", routeEntry.linkCount, routeEntry.angle))
end

local function onDrawDebug(focusPosition)
	if not route then return end
	
	local firstPos = route[1].pos + up
	local firstPosDistance = (firstPos - focusPosition):length()
	
	if firstPosDistance < debugDrawDistance then
		debugDrawer:drawText(firstPos:toPoint3F(), debugText(route[1]), textColor)
	end
	
	for i = 2, #route do
		local secondPos = route[i].pos + up
		local secondPosDistance = (secondPos - focusPosition):length()
		
		if firstPosDistance < debugDrawDistance or secondPosDistance < debugDrawDistance then
			debugDrawer:drawCylinder(firstPos:toPoint3F(), secondPos:toPoint3F(), 0.04, straightLineColor)
		end
		
		if secondPosDistance < debugDrawDistance then
			debugDrawer:drawText(secondPos:toPoint3F(), debugText(route[i]), textColor)
			
			local color
			
			if indexOfInterest(i) then
				color = angledLineColor
			else
				color = straightLineColor
			end
			
			debugDrawer:drawSphere(secondPos:toPoint3F(), 0.5, color)
		end
		
		firstPos = secondPos
		firstPosDistance = secondPosDistance
	end
end


----------------------------------------------------------------------------------------------------
-- Module interface
----------------------------------------------------------------------------------------------------

return {
	onPhysicsEngineEvent = onPhysicsEngineEvent,
	onScenarioChange = onScenarioChange,
	onRaceStart = onRaceStart,
	--onDrawDebug = onDrawDebug
}